<?php

declare(strict_types = 1);

namespace Drupal\brightcove\Services;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Serialization\SerializationInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\KeyValueStore\DatabaseStorageExpirable;

/**
 * Expirable access token storage.
 */
final class ExpirableAccessTokenStorage extends DatabaseStorageExpirable {

  /**
   * Time.
   */
  protected TimeInterface|string|null $time;

  /**
   * Initializes an expirable access token storage.
   *
   * @param \Drupal\Component\Serialization\SerializationInterface $serializer
   *   Serializer.
   * @param \Drupal\Core\Database\Connection $connection
   *   Database connection.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   Time.
   */
  public function __construct(
    SerializationInterface $serializer,
    Connection $connection,
    TimeInterface $time,
  ) {
    parent::__construct('brightcove_access_token', $serializer, $connection);

    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public function has($key): bool {
    try {
      return (bool) $this->connection->query('SELECT 1 FROM {' . $this->connection->escapeTable($this->table) . '} WHERE [collection] = :collection AND [name] = :key AND [expire] > :now', [
        ':collection' => $this->collection,
        ':key' => $key,
        ':now' => $this->time->getCurrentTime(),
      ])->fetchField();
    }
    catch (\Exception $e) {
      $this->catchException($e);
      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getAll(): array {
    try {
      $values = $this->connection->query(
        'SELECT [name], [value] FROM {' . $this->connection->escapeTable($this->table) . '} WHERE [collection] = :collection AND [expire] > :now',
        [
          ':collection' => $this->collection,
          ':now' => $this->time->getCurrentTime(),
        ])->fetchAllKeyed();
      return array_map([$this->serializer, 'decode'], $values);
    }
    catch (\Exception $e) {
      $this->catchException($e);
    }
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function getMultiple(array $keys): array {
    try {
      $values = $this->connection->query(
        'SELECT [name], [value] FROM {' . $this->connection->escapeTable($this->table) . '} WHERE [expire] > :now AND [name] IN ( :keys[] ) AND [collection] = :collection',
        [
          ':now' => $this->time->getCurrentTime(),
          ':keys[]' => $keys,
          ':collection' => $this->collection,
        ])->fetchAllKeyed();
      return array_map([$this->serializer, 'decode'], $values);
    }
    catch (\Exception $e) {
      $this->catchException($e);
    }
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function doSetWithExpire($key, $value, $expire): void {
    $this->connection->merge($this->table)
      ->keys([
        'name' => $key,
        'collection' => $this->collection,
      ])
      ->fields([
        'value' => $this->serializer->encode($value),
        'expire' => $this->time->getCurrentTime() + $expire,
      ])
      ->execute();
  }

}
